from inspect import getsource
from types import ModuleType
from typing import Callable, Optional

from ..base.template import MixinOperation
from ..base.toolchain import Container, DefaultToolChain, ToolChain
from ..base.typing import M, NameSpace


class Processor(MixinOperation[Callable[[str], str]]):
    """
    If you want to get the soure code of a method and use `split` and
    `replace` to modify and redefine it,Try `Processor`.

    When you try to use this, please make sure you configure
    the correct module of your target, or you can use `extra_namespace` to
    supplement the missing things.

    Don't try to use it with external code, like the code of cpython, it will crash.
    """

    module: Optional[ModuleType]
    extra_namespace: Optional[NameSpace]
    target_name: Optional[str]
    prefix_indent: Optional[int]

    def __init__(
        self,
        argument: Callable[[str], str],
        level: int = 1,
        target_name: Optional[str] = None,
        prefix_indent: Optional[int] = None,
        module: Optional[ModuleType] = None,
        extra_namespace: Optional[NameSpace] = None,
    ) -> None:
        super().__init__(argument, level)
        self.target_name = target_name
        self.prefix_indent = prefix_indent
        self.module = module
        self.extra_namespace = extra_namespace

    @staticmethod
    def configure(
        level: int = 1,
        target_name: Optional[str] = None,
        prefix_indent: Optional[int] = None,
        module: Optional[ModuleType] = None,
        extra_namespace: Optional[NameSpace] = None,
    ) -> Callable[[Callable[[str], str]], "Processor"]:
        return lambda argument: Processor(
            argument=argument,
            level=level,
            prefix_indent=prefix_indent,
            target_name=target_name,
            module=module,
            extra_namespace=extra_namespace,
        )

    def mixin(
        self,
        target: M,
        toolchain: ToolChain = DefaultToolChain,
    ) -> None:
        target_name = (
            self.target_name if self.target_name is not None else self.argument.__name__
        )

        return toolchain.tool_setattr(
            target,
            target_name,
            Container(
                None if self.module is None else self.module.__dict__,
                self.extra_namespace,
            ).define_function(
                function_name=target_name,
                indent=self.prefix_indent,
                source=self.argument(
                    getsource(
                        toolchain.tool_getattr(
                            target,
                            target_name,
                        )
                    )
                ),
            ),
        )



from typing import Callable, Optional

from ..base.template import MixinOperation
from ..base.toolchain import DefaultToolChain, ToolChain
from ..base.typing import M


class OverWrite(MixinOperation[Callable]):
    """
    OverWrite is rude and it will cover the target method.

    If the target method doesn't exist,
    overwrite will add overwrite method to target class.

    Try avoid using `OverWrite` with other `OverWrite`.
    """

    target_name: Optional[str]

    def __init__(
        self,
        argument: Callable,
        target_name: Optional[str] = None,
    ) -> None:
        super().__init__(argument)
        self.target_name = target_name

    @staticmethod
    def configure(
        target_name: Optional[str] = None,
    ) -> Callable[[Callable], "OverWrite"]:
        return lambda argument: OverWrite(
            argument=argument,
            target_name=target_name,
        )

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        target_name = (
            self.argument.__name__ if self.target_name is None else self.target_name
        )
        self.argument.__qualname__ = f"{target.__name__}.{target_name}"
        return toolchain.tool_setattr(
            target,
            target_name,
            self.argument,
        )



from abc import ABC
from typing import Any, Generic

from .toolchain import DefaultToolChain, ToolChain
from .typing import M, T


class MixinOperation(Generic[T], ABC):
    """The MixinOperation is the base of All Operation.

    The generic `MixinOperation` is the type of argument.

    `level` will affect to the mixin order, default `1`.

    If you call the `MixinOperation` or call the subclass of this,
    it will call the `MixinOperation.argument`
    """

    argument: T
    level: int

    def __init__(self, argument: T, level: int = 1) -> None:
        self.argument = argument
        self.level = level

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        raise NotImplementedError(
            f"Not Ready to use this Operation to modify '{target}' via '{toolchain}'"
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, Optional

from saleyo.base.typing import RT, P


class BroadCaster(Generic[P, RT], ABC):
    """Abstract Definition of `BroadCaster`"""

    def add_listener(
        self,
        listener: Callable[P, RT],
        key: Optional[Any] = None,
        key_rev: Optional[Callable[[Any], Any]] = None,
        disposable: bool = False,
    ):
        listeners = self.listeners(disposable)
        hashkey = key if key else hash(listener)
        if hashkey in listeners:
            raise Exception(f"Key: {key} Duplicate.")
        listeners[hashkey] = listener
        if key_rev:
            key_rev(hashkey)

    def remove_listener(self, key: Any, disposable: bool = False):
        listeners = self.listeners(disposable)

        if key not in listeners:
            raise Exception(f"{key} not in listeners.")

        del self.listeners(disposable)[key]

    def notifiers(
        self,
        disposable: bool = False,
        on_value: Optional[Callable[[RT], Any]] = None,
    ) -> Callable[P, None]:
        def notify(*args: P.args, **kwargs: P.kwargs):
            listeners = self.listeners(disposable)
            for key, listener in listeners.copy().items():
                value = listener(*args, **kwargs)
                if value and on_value:
                    on_value(value)

                if disposable:
                    del listeners[key]

        return notify

    def all_notifiers(
        self, on_value: Optional[Callable[[RT], Any]] = None
    ) -> Callable[P, None]:
        def notify(*args: P.args, **kwargs: P.kwargs):
            self.notifiers(False, on_value=on_value)(*args, **kwargs)
            self.notifiers(True, on_value=on_value)(*args, **kwargs)

        return notify

    def notifier(
        self,
        key: Any,
        disposable: bool = False,
        on_value: Optional[Callable[[RT], Any]] = None,
    ) -> Callable[P, None]:
        def notify(*args: P.args, **kwargs: P.kwargs):
            listeners = self.listeners(disposable)
            if key not in listeners:
                raise Exception(f"Cant Find {key} in listeners !")
            value = listeners[key](*args, **kwargs)
            if value and on_value:
                on_value(value)

            if disposable:
                del listeners[key]

        return notify

    @abstractmethod
    def listeners(self, disposable: bool = False) -> Dict[Any, Callable[P, RT]]: ...

    @abstractmethod
    def instance() -> "BroadCaster": ...

    @abstractmethod
    def initialize():
        raise NotImplementedError



from collections import OrderedDict
from types import ModuleType
from typing import Any, Callable, Dict, Optional, Union

from saleyo.base.broadcast import BroadCaster
from saleyo.base.typing import IterableOrSingle, M


class ImportBroadCaster(BroadCaster[[str, ModuleType], IterableOrSingle[M]]):
    _instance: Optional["ImportBroadCaster"] = None
    _initialize_flag: bool = False
    _listeners: OrderedDict[
        Union[str, int], Callable[[str, ModuleType], IterableOrSingle[M]]
    ] = OrderedDict()

    _disposable_listeners: OrderedDict[
        Union[str, int], Callable[[str, ModuleType], IterableOrSingle[M]]
    ] = OrderedDict()

    def listeners(
        self,
        disposable: bool = False,
    ) -> Dict[Any, Callable[[str, ModuleType], IterableOrSingle[M]]]:
        return self._disposable_listeners if disposable else self._listeners

    @staticmethod
    def instance() -> "ImportBroadCaster":
        if not ImportBroadCaster._instance:
            ImportBroadCaster._instance = ImportBroadCaster()
        return ImportBroadCaster._instance

    @staticmethod
    def initialize():
        if ImportBroadCaster._initialize_flag:
            return

        import sys

        class ImportBroadCastDict(type(sys.modules)):
            def __setitem__(self, key: str, value: ModuleType) -> None:
                ImportBroadCaster.instance().all_notifiers()(key, value)

                return super().__setitem__(key, value)

        sys.modules = ImportBroadCastDict(sys.modules)
        ImportBroadCaster._initialize_flag = True



from typing import Any

from ..base.template import MixinOperation
from ..base.toolchain import DefaultToolChain, ToolChain
from ..base.typing import M


class ReName(MixinOperation[str]):
    """
    Rename the target.
    """

    new: str

    def __init__(self, old: str, new: str, level=1) -> None:
        super().__init__(old, level)
        self.new = new

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        old = toolchain.tool_getattr(target, self.argument)
        toolchain.tool_delattr(target, self.argument)
        return toolchain.tool_setattr(target, self.new, old)


class Del(MixinOperation[str]):
    """
    Delete something named `argument` this from target
    """

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        return toolchain.tool_delattr(target, self.argument)


class Alias(MixinOperation[str]):
    """will copy the `argument` attribute to `alias`"""

    alias: str

    def __init__(self, argument: str, alias: str, level=1) -> None:
        super().__init__(argument, level)
        self.alias = alias

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        return toolchain.tool_setattr(
            target, self.alias, toolchain.tool_getattr(target, self.argument)
        )


class Insert(MixinOperation[Any]):
    """Insert a method to target, and will cover target when target exists."""

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        return toolchain.tool_setattr(target, self.argument.__name__, self.argument)



from saleyo import Intercept, InvokeEvent, Mixin, Post, Pre


class Foo:
    def demo(self):
        print("goodbye~")


@Mixin(target=Foo)
class MixinFoo:
    @Intercept.configure(target_name="demo")
    @staticmethod
    def intercept_demo(_: InvokeEvent):
        return InvokeEvent(lambda: print("hello world"))

    @Pre.configure(target_name="demo")
    def pre_demo(*arg):
        print("pre hello world")

    @Post.configure(target_name="demo")
    def post_demo(*arg):
        print("post hello world")


Foo().demo()



from typing import Any, Callable, Optional, Union

from saleyo.broadcast.compile import (
    CompileBoundary as CompileBoundary,
)
from saleyo.broadcast.compile import (
    CompileBroadCaster,
    CompileInfo,
)


class CompileTime:
    """
    This decorator is **dangerous**, it will affect the compile of .pyc.
    If you found it not work, please consider delete the target import `__py_cache__`
    and then use a `CompileBoundary`
    """

    locator: Callable[[CompileInfo], bool]
    dispose_key: Any
    key: Optional[Any]
    auto_dispose: bool
    disposable: bool

    def __init__(
        self,
        locator: Callable[[CompileInfo], bool],
        key: Optional[Any] = None,
        initialize: bool = True,
        auto_dispose: bool = True,
        disposable: bool = False,
    ) -> None:
        self.locator = locator
        self.key = key
        self.auto_dispose = auto_dispose
        self.disposable = disposable

        if initialize:
            CompileBroadCaster.instance().initialize()

    def __key_rev(self, key):
        self.dispose_key = key

    def __call__(self, processor: Callable[[CompileInfo], CompileInfo]):
        broadcast = CompileBroadCaster.instance()

        def listener(info: CompileInfo):
            if self.locator(info):
                if self.auto_dispose and self.dispose_key:
                    broadcast.remove_listener(self.dispose_key, self.disposable)

                return processor(info)

        broadcast.add_listener(
            listener=listener,
            key=self.key,
            key_rev=self.__key_rev if self.auto_dispose else None,
            disposable=self.disposable,
        )

        return processor


class CompileToken(CompileTime):
    """
    Alternative version of CompileTime, just modify source.
    But it is also **dangerous**.
    """

    def __call__(
        self, processor: Callable[[Union[str, bytes, Any]], Union[str, bytes, Any]]
    ):
        def wrapper(info: CompileInfo):
            return info.copy(source=processor(info.source))

        return super().__call__(wrapper)



"""
**Saleyo** is a package to modify external python code in runtime.

`base` module defined some abstract class/type and
you can extend Saleyo by this.

`broadcast` module defines some `BroadCaster`, the broadcaster initialize to inject
functions in builtins/importlib/... and broadcast to support some decorators work.

`decorator` module defines some Decorators, it's the most import module in Saleyo.

`operation` module defines some default `MixinOperation`.

Don't know how to start? Please see the part of Basic Tutorial in README.

The two links below are available.

https://github.com/H2Sxxa/saleyo/blob/main/README.md

https://pypi.org/project/saleyo/
"""

from . import base as base
from . import broadcast as broadcast
from . import decorator as decorator
from . import operation as operation
from .base.template import MixinOperation as MixinOperation
from .base.toolchain import Arguments as Arguments
from .base.toolchain import CPyToolChain as CPyToolChain
from .base.toolchain import GCToolChain as GCToolChain
from .base.toolchain import InvokeEvent as InvokeEvent
from .base.toolchain import ToolChain as ToolChain
from .decorator import Ancestor as Ancestor
from .decorator import CompileTime as CompileTime
from .decorator import CompileToken as CompileToken
from .decorator import FileLoadTime as FileLoadTime
from .decorator import Mixin as Mixin
from .operation import Accessor as Accessor
from .operation import Alias as Alias
from .operation import Both as Both
from .operation import Del as Del
from .operation import FunctionAccessor as FunctionAccessor
from .operation import Insert as Insert
from .operation import Intercept as Intercept
from .operation import OverWrite as OverWrite
from .operation import Post as Post
from .operation import Pre as Pre
from .operation import Processor as Processor
from .operation import ReName as ReName



from typing import Any, Callable, Optional

from saleyo.broadcast.file import FileLoaderBroadCaster


class FileLoadTime:
    """
    Get SourceCode(.py) / ByteCode(.pyc) from Module Loader and Modify it.

    Please use some third-party packages to modify it.

    When the target module has no cache, it has the same effects to `CompileToken`
    """

    locator: Callable[[str, bytes], bool]
    dispose_key: Any
    key: Optional[Any]
    auto_dispose: bool
    disposable: bool

    def __init__(
        self,
        locator: Callable[[str, bytes], bool],
        key: Optional[Any] = None,
        initialize: bool = True,
        auto_dispose: bool = True,
        disposable: bool = False,
    ) -> None:
        self.locator = locator
        self.key = key
        self.auto_dispose = auto_dispose
        self.disposable = disposable

        if initialize:
            FileLoaderBroadCaster.instance().initialize()

    def __key_rev(self, key):
        self.dispose_key = key

    def __call__(self, processor: Callable[[str, bytes], bytes]) -> Any:
        broadcast = FileLoaderBroadCaster.instance()

        def listener(path: str, bytecode: bytes):
            if self.locator(path, bytecode):
                if self.auto_dispose and self.dispose_key:
                    broadcast.remove_listener(self.dispose_key, self.disposable)

                return processor(path, bytecode)

        broadcast.add_listener(
            listener=listener,
            key=self.key,
            key_rev=self.__key_rev if self.auto_dispose else None,
            disposable=self.disposable,
        )

        return processor



from typing import Any, Generic, Type

from ..base.toolchain import DefaultToolChain, ToolChain
from ..base.typing import T


class Ancestor(Generic[T]):
    """
    Ancestor will add the `ancestor_class` to `target.__bases__`.
    (Please see `__call__` method)

    Please ensure the `target.__bases__` is not `(object,)`, this decorator requires
    the target has at least one super class.

    If `reverse`, the `argument` will add to the head of `target.__bases__`.

    Don't try to use it with external code and `module`, it will crash.
    """

    reverse: bool
    target: Type[Any]
    toolchain: ToolChain

    def __init__(
        self, target: Type[Any], toolchain: ToolChain = DefaultToolChain, reverse=False
    ) -> None:
        self.reverse = reverse
        self.target = target
        self.toolchain = toolchain

    def __call__(self, ancestor_class: Type[T]) -> Type[T]:
        assert self.target.__bases__ != (object,)

        self.toolchain.tool_setattr(
            self.target,
            "__bases__",
            (ancestor_class, *self.toolchain.tool_getattr(self.target, "__bases__"))
            if self.reverse
            else (
                *self.toolchain.tool_getattr(self.target, "__bases__"),
                ancestor_class,
            ),
        )

        return ancestor_class



from types import ModuleType
from typing import Any, Dict, Iterable, ParamSpec, Type, TypeVar, Union

# Generic

RT = TypeVar("RT")
"""
`RT` means `Return Type`
"""
T = TypeVar("T")
"""
`T` means `Type`
"""
P = ParamSpec("P")
"""
`P` means `Params`
"""

M = TypeVar("M", Type[Any], ModuleType, Any)
"""
These can be the target of mixin.
"""

# Alias

NameSpace = Dict[str, Any]
"""
`NameSpace` is the alias of `Dict[str, Any]`
"""

IterableOrSingle = Union[T, Iterable[T]]
"""
`IterableOrSingle[T]` is the alias of `Union[T, List[T]]` 
"""



from typing import Any

from saleyo import Accessor, Intercept, InvokeEvent, Mixin, OverWrite, Post, Pre


class Foo:
    __private = "private varible"

    def demo(self):
        pass


@Mixin(target=Foo)
class MixinFoo:
    # Will add a varible named `__private` to Foo and 
    # it has the same address with `_Foo__private`
    private: Accessor[str] = Accessor("__private")

    # Will Add the `func` to `Foo`
    @OverWrite
    def func(self):
        print("hello saleyo")

    # Will intercept the demo method and redirect to `lambda: print("hello world")`
    @Intercept.configure(target_name="demo")
    @staticmethod
    def intercept_demo(_: InvokeEvent):
        return InvokeEvent(lambda: print("hello world"))

    # Will call before `demo` call
    @Pre.configure(target_name="demo")
    def pre_demo(*arg):
        print("pre hello world")

    # Will call after `demo` call
    @Post.configure(target_name="demo")
    def post_demo(*arg):
        print("post hello world")


foo: Any = (
    Foo()
)  # Add the typing hint to avoid the error message from some IDE plugins.

print(foo.__private)  # Also `print(MixinFoo.private.value)`
foo.func()
foo.demo()



from saleyo import Alias, GCToolChain, Mixin

alias = Alias("upper", "do_upper")

Mixin(str, GCToolChain).apply_from_operations(alias)

print("Hello Saleyo!".do_upper()) # type: ignore
print("Hello Saleyo!".upper())



from typing import Any, Callable, Generic, Optional, ParamSpec

from ..base.template import MixinOperation
from ..base.toolchain import DefaultToolChain, InvokeEvent, ToolChain
from ..base.typing import M

_PA = ParamSpec("_PA")
_PB = ParamSpec("_PB")
_A = InvokeEvent[_PA, Any]
_B = InvokeEvent[_PB, Any]


class Intercept(Generic[_PA, _PB], MixinOperation[Callable[[_A[_PA]], _B[_PB]]]):
    """
    The `Intercept` allow you to intercept the arguments before invoking target method.

    Then, you can handle thefse arguments in your own function and make a redirect to
    another function.

    If you just want to modify the arguments or the result,
    please see `saleyo.operation.hook`.
    """

    target_name: Optional[str]

    def __init__(
        self,
        argument: Callable[[_A[_PA]], _B[_PB]],
        level: int = 1,
        target_name: Optional[str] = None,
    ) -> None:
        super().__init__(argument, level)
        self.target_name = target_name

    @staticmethod
    def configure(
        level: int = 1,
        target_name: Optional[str] = None,
    ) -> Callable[[Callable[[_A[_PA]], _B[_PB]]], "Intercept[_PA, _PB]"]:
        return lambda argument: Intercept(
            argument=argument,
            level=level,
            target_name=target_name,
        )

    def mixin(
        self,
        target: M,
        toolchain: ToolChain = DefaultToolChain,
    ) -> None:
        target_name = (
            self.target_name if self.target_name is not None else self.argument.__name__
        )

        original_function = toolchain.tool_getattr(
            target,
            target_name,
        )

        return toolchain.tool_setattr(
            target,
            target_name,
            lambda *args, **kwargs: self.argument(
                InvokeEvent(original_function, *args, **kwargs)
            ).invoke_target(),
        )



from typing import Any, Callable, Generic, Optional, Union

from ..base.template import MixinOperation
from ..base.toolchain import Arguments, DefaultToolChain, ToolChain
from ..base.typing import RT, M, P, T


class Post(MixinOperation[Callable[[T], Optional[RT]]]):
    """
    `Post` will call after the target method, and the callable should be decorated as
    `@staticmethod` and have one argument to receive the result of target method.

    If the `post` function return value is not `None`,
    it will replace the original result.
    """

    target_name: Optional[str]

    def __init__(
        self,
        argument: Callable[[T], Optional[RT]],
        target_name: Optional[str] = None,
        level: int = 1,
    ) -> None:
        super().__init__(argument, level)
        self.target_name = target_name

    @staticmethod
    def configure(
        level: int = 1,
        target_name: Optional[str] = None,
    ):
        return lambda argument: Post(
            argument=argument,
            target_name=target_name,
            level=level,
        )

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        target_name = (
            self.target_name if self.target_name is not None else self.argument.__name__
        )
        original_function: Callable[..., T] = toolchain.tool_getattr(
            target, target_name
        )

        def post(*args, **kwargs) -> Union[T, RT]:
            result = original_function(*args, **kwargs)
            post_result = self.argument(result)
            if post_result is not None:
                return post_result
            return result

        return toolchain.tool_setattr(target, target_name, post)


class Pre(Generic[P], MixinOperation[Callable[P, Optional[Arguments[P]]]]):
    """
    `Pre` will call before the target method, and the callable should be decorated as
    `@staticmethod` and have `*args,**kwargs` to receive the arguments of target method.

    If the `pre` function return value is a `Aruguments`(not `None`),
    it will replace the original arguments.
    """

    target_name: Optional[str]

    def __init__(
        self,
        argument: Callable[P, Optional[Arguments[P]]],
        target_name: Optional[str] = None,
        level: int = 1,
    ) -> None:
        super().__init__(argument, level)
        self.target_name = target_name

    @staticmethod
    def configure(
        level: int = 1,
        target_name: Optional[str] = None,
    ):
        return lambda argument: Pre(
            argument=argument,
            target_name=target_name,
            level=level,
        )

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        target_name = (
            self.target_name if self.target_name is not None else self.argument.__name__
        )
        original_function = toolchain.tool_getattr(target, target_name)

        def pre(*args: P.args, **kwargs: P.kwargs) -> Any:
            arguments = self.argument(*args, **kwargs)
            if arguments is not None:
                return original_function(*arguments.args, **arguments.kwargs)
            return original_function(*args, **kwargs)

        return toolchain.tool_setattr(target, target_name, pre)


class Both(
    Generic[P, RT],
    MixinOperation[Callable[[Arguments[P], Callable[P, RT]], RT]],
):
    """
    `Both` will call before the target method, and the callable should be decorated as
    `@staticmethod` and have `*args,**kwargs` to receive the arguments of target method.

    `Both` will give you the original target method and the arguments.
    """

    target_name: Optional[str]

    def __init__(
        self,
        argument: Callable[[Arguments[P], Callable[P, RT]], RT],
        target_name: Optional[str] = None,
        level: int = 1,
    ) -> None:
        super().__init__(argument, level)
        self.target_name = target_name

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        target_name = (
            self.target_name if self.target_name is not None else self.argument.__name__
        )
        original_function = toolchain.tool_getattr(target, target_name)

        def both(*args: P.args, **kwargs: P.kwargs) -> RT:
            return self.argument(Arguments(*args, **kwargs), original_function)

        return toolchain.tool_setattr(target, target_name, both)



from saleyo import Insert, Mixin
from saleyo.operation.accessor import Accessor, FunctionAccessor


class Foo:
    __value: int = 1

    def __increase_value(self) -> None:
        self.__value += 1

    def get_val(self):
        return self.__value


@Mixin(target=Foo)
class MixinFoo:
    value_accessor: Accessor[int] = Accessor("__value")
    increase_value: FunctionAccessor[[Foo], None] = FunctionAccessor("__increase_value")

    @Insert
    def helloworld(self):
        assert isinstance(self, Foo)
        MixinFoo.value_accessor.value += 1
        MixinFoo.increase_value(self)


foo: MixinFoo = Foo()
foo.helloworld()
assert isinstance(foo, Foo)
print(foo.get_val())
print(MixinFoo.value_accessor)



from types import ModuleType
from typing import Any, Callable, Generic, Iterable, List, Optional, Union

from ..base.template import MixinOperation
from ..base.toolchain import DefaultToolChain, ToolChain
from ..base.typing import IterableOrSingle, M, T


class Mixin(Generic[M]):
    """
    A `Mixin` Decorator is used to invoke all the `MixinOperation` in Mixin Class.

    If the target is not a common class, you should custom the toolchain yourself.

    **When the target if a `Iterable` object, Please use [`target`]**

    It is recommend to use `assert isinstance(self, <target>)` at the head of
    operation functions, although there may be some performance cost,
    but it is worth it in most conditions.

    Allow to have more than one target, but that's not recommended.
    """

    target: Iterable[M]
    toolchain: ToolChain
    reverse_level: bool

    def __init__(
        self,
        target: IterableOrSingle[M],
        toolchain: ToolChain = DefaultToolChain,
        reverse_level: bool = False,
    ) -> None:
        self.target = target if isinstance(target, Iterable) else [target]
        self.toolchain = toolchain
        self.reverse_level = reverse_level

    @staticmethod
    def from_name(
        target: IterableOrSingle[str],
        toolchain: ToolChain = DefaultToolChain,
        reverse_level: bool = False,
        qualname: bool = False,
    ) -> "Mixin":
        """
        Will find target classes via `object.__subclasses__()`.

        If want to find a class named `Foo`, default use the `Foo` to match,
        it will use `module.to.Foo` to match when `qualname` enabled.

        Please use this after the definition of target class.

        The method may takes lots of time when there are a whole lot classes,
        recommand to use `@Mixin()` directly if you can.
        """

        target = target if isinstance(target, Iterable) else [target]

        return Mixin(
            filter(
                lambda clazz: clazz.__name__ in target
                if qualname
                else clazz.__qualname__ in target,
                object.__subclasses__(),
            ),
            toolchain=toolchain,
            reverse_level=reverse_level,
        )

    @staticmethod
    def from_regex(
        pattern: str,
        pattern_flags: int = 0,
        toolchain: ToolChain = DefaultToolChain,
        reverse_level: bool = False,
        qualname: bool = False,
        full_match: bool = False,
    ) -> "Mixin":
        """
        Will use regex pattern to find target classes from the `object.__subclasses__()`

        The `pattern` will convert to a `Pattern[str]` via `re.complie` and
        you can provide flags in `pattern_flags`.

        If want to find a class named `Foo`, default use the `Foo` to match,
        it will use `module.to.Foo` to match when `qualname` enabled.

        Please use this after the definition of target class.

        The method may takes lots of time when there are a whole lot classes,
        recommand to use `@Mixin()` directly if you can.
        """
        import re

        regex_pattern = re.compile(pattern=pattern, flags=pattern_flags)
        matcher = regex_pattern.fullmatch if full_match else regex_pattern.match

        return Mixin(
            list(
                filter(
                    lambda clazz: matcher(
                        clazz.__name__ if qualname else clazz.__qualname__
                    )
                    is not None,
                    object.__subclasses__(),
                )
            ),
            toolchain=toolchain,
            reverse_level=reverse_level,
        )

    @staticmethod
    def lazy(
        locator: Callable[[str, ModuleType], Optional[IterableOrSingle[M]]],
        toolchain: ToolChain = DefaultToolChain,
        reverse_level: bool = False,
        key: Optional[Any] = None,
        initialize: bool = True,
        auto_dispose: bool = True,
        disposable: bool = False,
    ) -> Callable[[object], Union[int, str]]:
        """
        See `Mixin.lazy_mixin`
        """
        return lambda mixin: Mixin.lazy_mixin(
            mixin,
            locator,
            toolchain,
            reverse_level,
            key,
            initialize,
            auto_dispose,
            disposable,
        )

    @staticmethod
    def lazy_mixin(
        mixin: object,
        locator: Callable[[str, ModuleType], Optional[IterableOrSingle[M]]],
        toolchain: ToolChain = DefaultToolChain,
        reverse_level: bool = False,
        key: Optional[Any] = None,
        initialize: bool = True,
        auto_dispose: bool = True,
        disposable: bool = False,
    ) -> Union[int, str]:
        """
        Please call lazy before import a module.

        when the `factory` return a non-None value, the Mixin Class will apply to target

        You can add the listener yourself to controller more details,
        please see `saleyo.base.import_broadcast`

        initialize: initialize the import-broadcast

        key: specific the key of listener, you can remove it yourself

        auto_dispose: remove listener after first Mixin

        disposable: remove listener after importing first module
        """
        from saleyo.broadcast.importmod import ImportBroadCaster

        broadcast = ImportBroadCaster.instance()

        if initialize:
            broadcast.initialize()

        dispose_key = None

        def key_rev(key):
            nonlocal dispose_key
            dispose_key = key

        def listener(k, v):
            target = locator(k, v)
            if target:
                Mixin(
                    target=target, toolchain=toolchain, reverse_level=reverse_level
                ).apply_from_class(mixin)
                if dispose_key:
                    broadcast.remove_listener(dispose_key, disposable=disposable)

        broadcast.add_listener(
            listener,
            key,
            disposable=disposable,
            key_rev=key_rev if auto_dispose else None,
        )

        return key if key else hash(listener)

    def apply_from_class(self, mixin: T) -> T:
        """
        Recommand to use it from `@Mixin`
        """
        members: List[MixinOperation] = sorted(
            filter(
                lambda member: isinstance(member, MixinOperation),
                map(
                    lambda name: self.toolchain.tool_getattr(mixin, name),
                    filter(lambda name: not name.startswith("_"), dir(mixin)),
                ),
            ),
            key=lambda member: member.level,
            reverse=self.reverse_level,
        )
        for member in members:
            for target in self.target:
                member.mixin(target=target, toolchain=self.toolchain)
        return mixin

    def apply_from_operations(
        self, operations: IterableOrSingle[MixinOperation]
    ) -> None:
        """
        Use operations from a `IterableOrSingle[MixinOperation]` and apply to target.

        Always used to mixin class manually.

        ```python
        mixin = Mixin(...)
        op1 = ...
        op2 = ...
        mixin.apply_from_operations([op1, op2])
        ```
        """
        for operation in (
            operations if isinstance(operations, Iterable) else [operations]
        ):
            for target in self.target:
                operation.mixin(target=target, toolchain=self.toolchain)

    def __call__(self, mixin: T) -> Union[M, T]:
        return self.apply_from_class(mixin=mixin)



from saleyo.base.import_broadcast import activate as activate
from saleyo.base.import_broadcast import add_listen_import

add_listen_import(lambda k, v: print(k, v))


def something():
    import targetmod as targetmod


something()



from collections import OrderedDict
from typing import Any, Callable, Dict, Optional, Union

from saleyo.base.broadcast import BroadCaster
from saleyo.base.toolchain import Arguments


class FileLoaderBroadCaster(BroadCaster[[str, bytes], Optional[bytes]]):
    _instance: Optional["FileLoaderBroadCaster"] = None
    _listeners: OrderedDict[
        Union[str, int], Callable[[str, bytes], Optional[bytes]]
    ] = OrderedDict()

    _disposable_listeners: OrderedDict[
        Union[str, int], Callable[[str, bytes], Optional[bytes]]
    ] = OrderedDict()

    @staticmethod
    def instance() -> "FileLoaderBroadCaster":
        if not FileLoaderBroadCaster._instance:
            FileLoaderBroadCaster._instance = FileLoaderBroadCaster()
        return FileLoaderBroadCaster._instance

    def listeners(
        self, disposable: bool = False
    ) -> Dict[Any, Callable[[str, bytes], Optional[bytes]]]:
        return self._disposable_listeners if disposable else self._listeners

    @staticmethod
    def initialize():
        from importlib._bootstrap_external import FileLoader

        from saleyo.operation import Both

        instance = FileLoaderBroadCaster.instance()

        def get_data(origin: Arguments[FileLoader, str], func):
            bytecode = func(*origin.args, **origin.kwargs)
            arg_path: str = ""

            def on_value(rev: Optional[bytes]):
                if rev:
                    nonlocal bytecode
                    bytecode = rev

            def fix_arguments(self: FileLoader, path: str):
                nonlocal arg_path
                arg_path = path

            fix_arguments(*origin.args, **origin.kwargs)
            instance.all_notifiers(on_value)(arg_path, bytecode)

            return bytecode

        Both(get_data).mixin(FileLoader)



"""
The `operations` always inner a class and call by `@Mixin` decorator.
"""

from .accessor import Accessor as Accessor
from .accessor import FunctionAccessor as FunctionAccessor
from .hook import Both as Both
from .hook import Post as Post
from .hook import Pre as Pre
from .intercept import Intercept as Intercept
from .modify import Alias as Alias
from .modify import Del as Del
from .modify import Insert as Insert
from .modify import ReName as ReName
from .overwrite import OverWrite as OverWrite
from .processor import Processor as Processor



from saleyo.decorator.file import FileLoadTime


@FileLoadTime(lambda path, bytecode: True)
def mixin(path, bytecode):
    print(path)
    print(bytecode)
    return bytecode


import targetmodule  # noqa: E402, F401



from typing import Any, Union

import targetmodule

from saleyo.decorator.compile import CompileBoundary, CompileToken


@CompileToken(lambda info: "targetmodule.py" in str(info.filename))
def mixin(token: Union[str, bytes, Any]):
    if not isinstance(token, bytes):
        return
    return token.replace(b"static", b"bye")


CompileBoundary.recompile_module(targetmodule)


print(targetmodule.Static().FIELD)  # hello bye



from types import ModuleType
from typing import Any

from saleyo.broadcast.importmod import ImportBroadCaster
from saleyo.decorator.mixin import Mixin
from saleyo.operation.hook import Pre

broadcast = ImportBroadCaster.instance()


def locator(name: str, module: ModuleType):
    if name == "targetmod" and module.__dict__.__contains__("NeedLazyMixin"):
        return module.NeedMixin
    return None


@Mixin.lazy(locator)
class MixinTarget:
    @Pre
    @staticmethod
    def hello(this: Any) -> None:
        print(this)
        print("Pre Hook")


print(broadcast.listeners())
import targetmod  # noqa: E402

print(broadcast.listeners())

targetmod.NeedMixin().hello()



import test_module_a
from saleyo.decorator.mixin import Mixin
from saleyo.operation.overwrite import OverWrite


@Mixin(test_module_a)
class MixinModuleA:
    @OverWrite.configure(target_name="hello")
    def hello_over() -> None:
        print("goodbye")




from ctypes import POINTER as _POINTER
from ctypes import cast as _cast
from ctypes import py_object as _py_object
from dataclasses import dataclass
from gc import get_referents as _get_referents
from types import FunctionType
from typing import Any, Callable, Dict, Generic, Optional

from ..base.typing import NameSpace, P, T


@dataclass
class ToolChain:
    """
    The tool class to do mixin.

    Default to use `getattr`/`setattr`/`hasattr`/`delattr`.
    """

    tool_getattr: Callable[[Any, str], Any] = getattr
    tool_setattr: Callable[[Any, str, Any], None] = setattr
    tool_hasattr: Callable[[Any, str], bool] = hasattr
    tool_delattr: Callable[[Any, str], None] = delattr


DefaultToolChain = ToolChain()

GCToolChain = ToolChain(
    tool_getattr=lambda _object, _name: _get_referents(_object.__dict__)[0][_name],
    tool_hasattr=lambda _object, _name: _name in _get_referents(_object.__dict__)[0],
    tool_setattr=lambda _object, _name, _attr: _get_referents(_object.__dict__)[
        0
    ].update({_name: _attr}),
    tool_delattr=lambda _object, _name: _get_referents(_object.__dict__)[0].__delitem__(
        _name
    ),
)
"""
GC ToolChain use the `get_referents` functions in `gc` and 
it can modify some special class.

Notice: There is no guarantee that it can modify any class,
and this method is rude and dangerous, avoid using it in produce environment.
"""


def _cpy_get_dict(_object: Any) -> Dict[str, Any]:
    return _cast(
        id(_object) + type(_object).__dictoffset__, _POINTER(_py_object)
    ).contents.value


CPyToolChain = ToolChain(
    tool_getattr=lambda _object, _name: _cpy_get_dict(_object)[_name],
    tool_hasattr=lambda _object, _name: _name in _cpy_get_dict(_object),
    tool_setattr=lambda _object, _name, _attr: _cpy_get_dict(_object).update(
        {_name: _attr}
    ),
    tool_delattr=lambda _object, _name: _cpy_get_dict(_object).__delitem__(_name),
)
"""
`CPyToolChain` use the `ctypes` to modify class, 
it's dangerous than other default toolchains.

Notice: There is no guarantee that it can modify any class, 
and this method is rude and dangerous, avoid using it in produce environment.
"""


class Container:
    """
    Container is A Class to keep a namespace and
    use this namespace to define function / variable / ...
    """

    environment: NameSpace

    def __init__(self, *namespace: Optional[NameSpace]) -> None:
        self.environment = {
            k: v for _ in namespace if _ is not None for k, v in _.items()
        }

    def exec(self, source: str) -> NameSpace:
        exec(source, self.environment)
        return self.environment

    def define_function(
        self, function_name: str, source: str, indent: Optional[int] = None
    ) -> FunctionType:
        if indent is None:
            indent = len(source[: source.find(source.lstrip()[0])])

        return self.exec(
            source="\n".join(
                [line.removeprefix(" " * indent) for line in source.splitlines()]
            )
        )[function_name]


class Arguments(Generic[P]):
    """
    `Argument` is used to call function,
    the Generic `P` is the params of target function.
    """

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"Arugument( positional: {self.args}, keyword: {self.kwargs} )"


class InvokeEvent(Generic[P, T]):
    """
    A `InvokeEvent` includes the target function and the arguments to
    call this functions.
    """

    target: Callable[P, T]
    argument: Arguments[P]

    def __init__(
        self,
        target: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__()
        self.target = target
        self.argument = Arguments(*args, **kwargs)

    def invoke(self, target: Callable[P, T]) -> T:
        return target(*self.argument.args, **self.argument.kwargs)

    def invoke_target(self) -> T:
        return self.target(*self.argument.args, **self.argument.kwargs)



from typing import Callable, Generic, Optional

from ..base.template import MixinOperation
from ..base.toolchain import DefaultToolChain, ToolChain
from ..base.typing import RT, M, P, T


class Accessor(Generic[T], MixinOperation[str]):
    """
    The Generic `T` is the type of target varible,
    if you want to visit a private function, try to use the subclass `FunctionAccessor`.

    Notice: The value only available after invoking the `mixin` method.

    If the `private` is `True`, will add target class name (like `_Foo`) to the prefix
    to argument, if the target is complex,
    you can set `private` to `False` and provide the true name by yourself.

    Also a variable named `argument` will add to target classes when `private` is `True`

    If you use `@Mixin` and have more than one target classes, the `value` will always
    be the varible of latest target.
    """

    _inner: Optional[T]
    _private: bool

    def __init__(self, argument: str, level=1, private=True) -> None:
        super().__init__(argument, level)
        self._inner = None
        self._private = private

    def mixin(self, target: M, toolchain: ToolChain = DefaultToolChain) -> None:
        self._inner = toolchain.tool_getattr(
            target,
            f"_{target.__name__}{self.argument}" if self._private else self.argument,
        )
        if self._private:
            toolchain.tool_setattr(
                target,
                self.argument,
                self._inner,
            )

    @property
    def value(self) -> T:
        """
        Don't use until The `mixin` method call.
        """
        assert self._inner
        return self._inner

    @value.setter
    def value(self, value: T):
        self._inner = value

    @value.deleter
    def value(self):
        # may cause some problems
        del self.value

    def __str__(self) -> str:
        return f"Accessor {{ value: {self._inner} ({id(self._inner)}) }}"


class FunctionAccessor(Generic[P, RT], Accessor[Callable[P, RT]]):
    """
    `FunctionAccessor` can be call directly.

    It's recommend to provide the Generic `P` and `T`, it can be useful in `__call__`.

    If the `private` is `True`, will add target class name (like `_Foo`) to the prefix
    to argument, if the target is complex, you can set `private` to `False` and
    provide the true name by yourself.

    Also a variable named `argument` will add to target classes when `private` is `True`

    If you just call in operation functions, you can just use a simple variable with
    `Callable[P, T]` type.

    If you use `@Mixin` and have more than one target classes, the `value` will always
    be the varible of latest target.

    ```python
    something: FunctionAccessor[[str], None] = FunctionAccessor("something")
    
    ```
    """

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> RT:
        """
        Call the `argument` function in target, don't use until The `mixin` method call.
        """
        assert self._inner
        return self._inner(*args, **kwargs)



"""The `decorator` should be used to decorate a class or function."""

from .ancestor import Ancestor as Ancestor
from .compile import CompileInfo as CompileInfo
from .compile import CompileTime as CompileTime
from .compile import CompileToken as CompileToken
from .file import FileLoadTime as FileLoadTime
from .mixin import Mixin as Mixin



import sys
from collections import OrderedDict
from dataclasses import dataclass
from importlib import reload
from os import unlink
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, Optional, Tuple, Union

from saleyo.base.broadcast import BroadCaster


@dataclass
class CompileInfo:
    """Concrete Arguments passed to `builtins.compile`"""

    source: Union[str, bytes, Any]
    filename: Union[str, bytes, Any]
    mode: str
    args: Tuple[Any]
    kwargs: Dict[str, Any]

    def copy(
        self,
        source: Optional[Union[str, bytes, Any]] = None,
        filename: Optional[Union[str, bytes, Any]] = None,
        mode: Optional[str] = None,
        args: Optional[Tuple[Any]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> "CompileInfo":
        return CompileInfo(
            source=source if source else self.source,
            filename=filename if filename else self.filename,
            mode=mode if mode else self.mode,
            args=args if args else self.args,
            kwargs=kwargs if kwargs else self.kwargs.copy(),
        )


class CompileBroadCaster(BroadCaster[[CompileInfo], Optional[CompileInfo]]):
    """BroadCast when `builtins.compile` called"""

    _instance: Optional["CompileBroadCaster"] = None
    _initialize_flag: bool = False
    _listeners: OrderedDict[
        Union[str, int], Callable[[CompileInfo], Optional[CompileInfo]]
    ] = OrderedDict()

    _disposable_listeners: OrderedDict[
        Union[str, int], Callable[[CompileInfo], Optional[CompileInfo]]
    ] = OrderedDict()

    def listeners(
        self,
        disposable: bool = False,
    ) -> Dict[Any, Callable[[CompileInfo], CompileInfo | None]]:
        return self._disposable_listeners if disposable else self._listeners

    @staticmethod
    def instance() -> "CompileBroadCaster":
        if not CompileBroadCaster._instance:
            CompileBroadCaster._instance = CompileBroadCaster()
        return CompileBroadCaster._instance

    @staticmethod
    def initialize():
        if CompileBroadCaster._initialize_flag:
            return

        import builtins

        origin_compile = compile

        def broadcast(
            source: Union[str, bytes, Any],
            filename: Union[str, bytes, Any],
            mode,
            *args,
            **kwargs,
        ):
            info = CompileInfo(source, filename, mode, args, kwargs)

            def on_value(rev: Optional[CompileInfo]):
                if rev:
                    nonlocal info
                    info = rev

            CompileBroadCaster.instance().all_notifiers(on_value=on_value)(info)

            return origin_compile(
                info.source,
                info.filename,
                info.mode,
                *info.args,
                **info.kwargs,
            )

        builtins.compile = broadcast

        CompileBroadCaster._initialize_flag = True


class CompileBoundary:
    """
    Module Inside will force compile.

    ## Example

    ```python
    import module_top

    # something

    with CompileBoundary() as compile:
        import module_a
        import module_b
        import module_c
        compile.recompile_module(module_top)
    ```
    """

    origin: bool
    no_cache: bool

    def __init__(self, no_cache: bool = True) -> None:
        self.origin = sys.dont_write_bytecode
        self.no_cache = no_cache

    def __enter__(self):
        if self.no_cache:
            self.activate()
        return self

    def __exit__(self, *_):
        if self.no_cache:
            self.deactivate()

    def activate(self):
        sys.dont_write_bytecode = True

    def deactivate(self):
        sys.dont_write_bytecode = self.origin

    @staticmethod
    def recompile_module(module: ModuleType):
        """
        When the module has been loaded before, recompile it to modify

        ## Example

        ```python
        import targetmodule

        # Mixin Here...

        with CompileBoundary(no_cache=False) as compile:
            compile.recompile_module(targetmodule)
        ```
        """
        if hasattr(module, "__cached__"):
            file = Path(module.__cached__)
            if file.exists():
                unlink(module.__cached__)
        reload(module)



from typing import Any, Type

from saleyo.base.template import MixinOperation
from saleyo.base.toolchain import ToolChain


class MyOperation(MixinOperation[Any]):
    def mixin(self, target: Type, toolchain: ToolChain = ...) -> None: ...



from saleyo import Arguments, GCToolChain, Mixin, Pre


@Mixin(target=str, toolchain=GCToolChain)
class MixinStr:
    @Pre.configure(target_name="format")
    def pre_format(self, *args, **kwargs) -> Arguments[...]:
        print(f"input args: {args}")
        print(f"input kwargs: {kwargs}")
        return Arguments(self, *args, **kwargs)


print("hello world {},{}".format("saleyo", "some"))
